function tnetTrainWeb_resnet_clean(varargin)
%FNCTRAIN Train FCN model using MatConvNet

	setupPaths();

	opts.resultDir  = '/dev/null'; % sprintf('/data/david/tnet-01-old-01lr-stiefel') ; 
	opts.imdbPath = '/dev/null';

	opts.showImages = true;
	opts.train = {};
	opts.net = {};
	opts.GPU = -1;
	opts.flip = true;
	opts.nPos = 2e3;   opts.nNeg = 2e3;
	opts.nPosVal = 20; opts.nNegVal = 10;
	opts.randSample = true;
	opts.seed = 0;
	opts.obClasses = [];
	opts.imageDir = '';
	opts.debugShow = 0; 
	opts = vl_argparse(opts,varargin);

	opts.train{end+1} = 'netName';
	opts.train{end+1} = opts.resultDir;
	opts = setupTrainOpts(opts);
	opts.train.extractStatsFn = @extractStats; 

	vl_xmkdir(opts.resultDir);

	rng(opts.seed,'twister');

	if exist(opts.imdbPath)~=2
		% download ilsvrc imdb
		url_    = 'http://www.robots.ox.ac.uk/~david/anchornet/imdb.mat';
		fprintf('fetching imdb from %s to %s\n',url_,opts.imdbPath);
		urlwrite(url_,opts.imdbPath);
		fprintf('done\n');
	end

	imdb = load(opts.imdbPath); 
	if isfield(imdb,'imdb')
		imdb = imdb.imdb;
	end

	if ~isempty(opts.imageDir)
		imdb.imageDir = opts.imageDir;
	end

	[~,okclsid_]          = ismember(opts.obClasses,imdb.classes.namePascal);
	[net,derOutputs]      = initializeAnchorNet( opts.net{:}, 'obClasses', okclsid_ ); 
	opts.train.derOutputs = derOutputs; % <------ !!!!!!!!!

	% show network
	% if true
	%   str = net.print({'data',[net.meta.normalization.imageSize(1:2) 3 1]},'format','dot');
	% 	displayDot(str,fullfile(opts.resultDir,'netDiag'));
	% end  

	rng(opts.seed,'twister'); 

	[imdb,TRAIN,VAL] = setupDataset(opts,imdb); 

	saveImdbPath = fullfile(opts.resultDir,'imdb.mat');
	save(saveImdbPath,'TRAIN','VAL','imdb');
	saveOptsPath = fullfile(opts.resultDir,'opts.mat');
	save(saveOptsPath,'opts');

	opts.train.gpus = opts.GPU; 

	bopts = setupBOpts(opts,net);

	precVars = {'keyPointModels','keyPointReluMasked', 'keyPointRelu','label','labelSum'};
	for precVar = precVars
		varidx = net.getVarIndex(precVar{1});
		if ~isnan(varidx)
			net.vars(varidx).precious = true ;	 
		end
	end 

	% Launch SGD
	% net.conserveMemory = false;
	info = cnn_train_dag_anchornet(net, imdb, getBatchWrapper(bopts), opts.train, ...
							  'train', TRAIN, ...
							  'val', VAL, ...
							  'extractStatsFn', extractStatsWrapper()) ; 

end

function clsmap = getClassMap(imdb)

	clsmap = zeros(20,1000);
	for cls = 1:20
		okin = unique(imdb.images.label(imdb.images.labelPascal==cls));
		clsmap(cls,okin) = 1;
		fprintf('%s:\n',imdb.classes.namePascal{cls});
		fprintf('   ');
		fprintf(' %s,',imdb.classes.description{okin});
		fprintf('\n');
	end

end

function opts = setupTrainOpts(opts,derOutputs)

	trainO = struct();
	trainO.gpus = 0 ;
	trainO.learningRate = NaN;
	trainO.batchSize = NaN;
	trainO.numSubBatches = NaN;
	trainO.continue = false;
	trainO.expDir = opts.resultDir ; 
	trainO.derOutputs = {'nan',nan};
	trainO.netName = '';
	trainO.saveProgress = true;
	trainO.debugShow = opts.debugShow; 

	opts.train = vl_argparse(trainO,opts.train);
	opts.train.numEpochs = numel(opts.train.learningRate) ;

end

function [imdb,train,val] = setupDataset(opts,imdb)
	
	imdb.negClass = NaN;
		
	% IMPORTANT !!!!
	imdb.images.labelINet = imdb.images.label;
	imdb.images.label = imdb.images.labelPascal;
	imdb.images.class = imdb.images.label;
	imdb.classes.name = imdb.classes.namePascal;
	% !!!!!!!!!!!!!!

	imdb.images.class(imdb.images.labelINet==0) = -1;
	[~,okclsid] = ismember(opts.obClasses,imdb.classes.namePascal);
	assert(numel(okclsid)>0);

	for clsi=1:numel(okclsid)
		fprintf(' pas class = %s; clsid = %d\n',opts.obClasses{clsi},okclsid(clsi));
	end

	oknegval = ismember(imdb.images.labelPascal,0) & imdb.images.set==2;
	val      = vl_colsubset(find(oknegval),opts.nNegVal);
	for clsid = okclsid
		okposval = ismember(imdb.images.labelPascal,clsid) & imdb.images.set==2;
		val2     = vl_colsubset(find(okposval), ceil(opts.nPosVal/numel(okclsid)) );
		val      = [val val2];
	end
	train = find(imdb.images.set==1 & ismember(imdb.images.labelPascal,[okclsid 0]) );
	assert(isempty(intersect(train,val)));

	assert(all(imdb.images.set(train)==1));
	assert(all(imdb.images.set(val)==2));

	imdb.images.set(:) = -1;
	imdb.images.set( train ) = 1;
	imdb.images.set( val   ) = 2;

	nPos = opts.nPos; nNeg = opts.nNeg;
	pos_  = find(imdb.images.set==1 & imdb.images.label ~= 0 );
	train = randSel(pos_,nPos);
	neg_  = find(imdb.images.set==1 & imdb.images.label == 0 );
	train = [ train, randSel(neg_,nNeg) ];


end

function [xs,idx] = randSel(x,n)

	idx = randi([1 numel(x)],1,n);
	xs  = x(idx);

end

function bopts = setupBOpts(opts,net)

	% Setup data fetching options
	bopts.rgbMean     = net.meta.normalization.averageImage ;
	bopts.useGpu      = numel(opts.train.gpus) > 0 ; 
	bopts.imageSize   = [net.meta.normalization.imageSize];
	bopts.flip        = opts.flip;
	bopts.randSample  = opts.randSample;
	bopts.nKeypoints  = net.meta.nKeypoints;
	bopts.obClassMask = net.meta.obClassMask;
	bopts.unfoldLabel = false;
	bopts.normCorr    = true;
	if net.meta.useMask
		vsz = net.getVarSizes({'data',bopts.imageSize});
		outsz = vsz{net.getVarIndex('keyPointRelu')};
		bopts.mask = outsz  ;
	end

end

% -------------------------------------------------------------------------
function fn = extractStatsWrapper(clean)
% -------------------------------------------------------------------------

	fn = @(net,state) extractStats(net,state) ;

end


function [crossCorr,autoCorr] = computeCorrs(net,layerName)

	li = net.getLayerIndex(layerName);

	if isnan(li)
		crossCorr = []; autoCorr = [];
		return;
	end

	layer  = net.layers(li);

	if isempty(layer.paramIndexes)
		crossCorr = []; autoCorr = [];
		return;
	end

	kpprms = net.params(layer.paramIndexes(1)).value; 
	kpder = net.params(layer.paramIndexes(1)).der; 

	sz     = [size(kpprms,1) size(kpprms,2) size(kpprms,3) size(kpprms,4)];

	if sz(4)==1
		crossCorr = []; autoCorr = [];
		return;
	end


	kpprms = reshape(kpprms, sz(1)*sz(2)*sz(3), sz(4) );
	
	switch class( layer.block )
		case 'dagnn.Conv'
			kpprms = bsxfun(@rdivide,kpprms, sqrt(sum(kpprms.*kpprms,1)));
			wht_ = abs(squeeze(kpprms)'*squeeze(kpprms)) ;
		case 'dagnn.ConvWhiten'
			if layer.block.nrm
				kpprms = bsxfun(@rdivide,kpprms, sqrt(sum(kpprms.*kpprms,1)));
			end
			p = layer.block.p;
			wht_ = abs(squeeze(kpprms)'*squeeze(kpprms)).^p ;
		otherwise
			error();
	end

	autoCorr = gather(mean(diag(wht_)));
	offdiagsel = find(~logical(diag(size(wht_))));
	crossCorr = mean(wht_(offdiagsel));

end


% -------------------------------------------------------------------------
function fn = getBatchWrapper(opts) 
% -------------------------------------------------------------------------

	fn = @(imdb,batch) getBatchAnchorNet(imdb,batch,opts) ;

end

function stats = extractStats(net,state)

	kpn = 'keyPointModels';
	[crossCorr,autoCorr] = computeCorrs(net,kpn);
	if ~isempty(crossCorr)
		stats.(sprintf('cc')) = gather(crossCorr);
		stats.(sprintf('ac')) = gather(autoCorr);
	end

	enc = 'shareEncoder'; 
	[crossCorr,autoCorr] = computeCorrs(net,enc);
	if ~isempty(crossCorr)  
		stats.(sprintf('ccEnc')) = gather(crossCorr);
		stats.(sprintf('acEnc')) = gather(autoCorr);
		val_ = net.params(net.getParamIndex([enc '_f']));
	end


	for li = 1:numel(net.layers)
		lnm = net.layers(li).name; 
		if ~isempty(strfind(lnm,'LOSS_'))
			stats.(lnm)   = gather(net.layers(li).block.average);
		end
	end

	kpr = net.getVarIndex('keyPointRelu');
	if ~isnan(kpr)
		activity = net.vars(kpr).value;
		lab_     = net.vars(net.getVarIndex('labelSum')).value;
		ok       = find(lab_>0);    
		posact   = activity(:,:,:,ok);
		actMean  = sum(abs(posact(:)))/numel(posact);
		stats.actMean = gather(actMean); 
		maxfilt  = max(max(activity,[],2),[],1);
		posact   = maxfilt(:,:,:,ok);
		actMean  = sum(abs(posact(:)))/numel(posact);
		stats.actMaxMean = gather(actMean);  
	end

end

function val = gatherVar(net,varnm)

	vari = net.getVarIndex(varnm);
	if isnan(vari)
		val = [];
	else
		val  = gather(net.vars(vari).value) ;
	end

end

% -------------------------------------------------------------------------
function epoch = findLastCheckpoint(modelDir)
% -------------------------------------------------------------------------
	list = dir(fullfile(modelDir, 'net-epoch-*.mat')) ;
	tokens = regexp({list.name}, 'net-epoch-([\d]+).mat', 'tokens') ;
	epoch = cellfun(@(x) sscanf(x{1}{1}, '%d'), tokens) ;
	epoch = max([epoch 0]) ;

end

% -------------------------------------------------------------------------
function [net, stats] = loadState(fileName)
% -------------------------------------------------------------------------
load(fileName, 'net', 'stats') ;
net = dagnn.DagNN.loadobj(net) ;

end